#include <ApplicationServices/ApplicationServices.h>
#include <Accelerate/Accelerate.h>

int main (int argc, const char * argv[]) {

	CMProfileRef	rgbRef    = NULL;
	CMProfileRef	labRef    = NULL;
	CMProfileRef	xyzRef    = NULL;
	CMWorldRef		cwRGBref  = NULL;
	CMWorldRef		cwXYZref  = NULL;
	CMError			err;
	
	// RGB profile: 
	// argv[1] or "./testRGB.icc" or system Profile
	// Whitepoint should be D50 and TRCs must contain 
	// only one entry (i.e. simple gamma, no curve)
	// testRGB.icc is a RGB profile with NTSC 1953 primaries, 
	// D50 whitepoint and Gamma 1.8
	CMProfileLocation pl;
	pl.locType = cmPathBasedProfile;
	if (argc <2)
		strcpy (pl.u.pathLoc.path, "testRGB.icc");
	else
		strcpy (pl.u.pathLoc.path, argv[1]);
		
	
	char*		aName; // RGB profile name
	UInt32		aCount;
	
	int			i, j; 
	int	clipped;    // out of RGB gamut
	int	clippedLab; // out of Uint16 Lab range
	
	// Lab FloatBitmap 
	float		lab[3] = {0.0, 0.0, 0.0};
	CMFloatBitmap map_lab = { 0,	// version
		{lab, lab+1, lab+2},		// base addrs of L,a,b
		1, 1,						// height, width
		3,							// rowStride
		3,							// colStride
		cmLabData,
		kCMFloatBitmapFlagsNone};

	// 2nd Lab FloatBitmap 
	float		lab2[3] = {0.0, 0.0, 0.0};
	CMFloatBitmap map_lab2 = { 0,	// version
		{lab2, lab2+1, lab2+2},		// base addrs of L,a,b
		1, 1,						// height, width
		3,							// rowStride
		3,							// colStride
		cmLabData,
		kCMFloatBitmapFlagsNone};

	// lab clipped to UInt16 range:
	float		labc[3] = {0.0, 0.0, 0.0};

	// RGB FloatBitmap 
	float		rgb[3] = {0.0, 0.0, 0.0};
	CMFloatBitmap map_rgb = { 0,	// version
		{rgb, rgb+1, rgb+2},		// base addrs of r,g,b
		1, 1,						// height, width
		3,							// rowStride
		3,							// colStride
		cmRGBData,
		kCMFloatBitmapFlagsNone};
	   	
	// XYZ FloatBitmap 
	float		xyz[3] = {0.0, 0.0, 0.0};
	CMFloatBitmap map_xyz = { 0,// version
		{xyz, xyz+1, xyz+2},			// base addrs of X,Y,Z
		1, 1,					// height, width
		3,						// rowStride
		3,						// colStride
		cmXYZData,
		kCMFloatBitmapFlagsNone};
		
	// 2nd XYZ FloatBitmap 
	float		xyz2[3] = {-1.0, -1.0, -1.0};
	CMFloatBitmap map_xyz2 = { 0,// version
		{xyz2, xyz2+1, xyz2+2},	 // base addrs of X,Y,Z
		1, 1,					 // height, width
		3,						 // rowStride
		3,						 // colStride
		cmXYZData,
		kCMFloatBitmapFlagsNone};

	// Fixed Lab 
	CMColor fixedColor[1];
	fixedColor[0].Lab.L = (UInt16) (lab[0] * 655.35);
	fixedColor[0].Lab.a = (UInt16)((lab[1] + 128.0) * 256); 
	fixedColor[0].Lab.b = (UInt16)((lab[2] + 128.0) * 256);
	
	// CMGetProfileElement params
	UInt32 size;
	void *dataP;
	unsigned char* dataPP;
	UInt32 count;
	float Rgamma, Ggamma, Bgamma;
	CMXYZType    fixedxyz;
	
	// clapack sgetrf and sgetrs params
	//The number of rows of the matrix A.  M >= 0.
	__CLPK_integer  m = 3;
	//The number of columns of the matrix A.  N >= 0.
	__CLPK_integer  n = 3;
	__CLPK_real a[m*n];
	// The leading dimension of  the  array  A.   LDA  >= max(1,M).
	__CLPK_integer lda=3;
	//  The pivot indices
	__CLPK_integer ipiv[3];
	__CLPK_integer info;
	// 	transpose 
	char trans = 'N';
	//  number of right hand sides
	__CLPK_integer nrhs = 1;
	// On entry, the right hand side matrix B.  
	// On  exit, the solution matrix X.
	__CLPK_real b[3];
	// leading dimension of  the  array  B
	__CLPK_integer ldb=3;
	
	// copy of matrix A
	float a2[m*n];

	// open RGB profile (argv[1] or "./testRGB.icc")
	err = CMOpenProfile (&rgbRef, &pl);
	if (err != noErr) {
		// open RGB profile failed
		// use system profile instead
		CMCloseProfile(rgbRef);
		err = CMGetSystemProfile(&rgbRef);
	}
	
	if (err == noErr) {
		// print profile name
		if (CMGetProfileDescriptions(rgbRef,NULL, &aCount, NULL, NULL, NULL, NULL) == noErr)
			aName = malloc(aCount);
		if (CMGetProfileDescriptions(rgbRef, aName, &aCount, NULL, NULL, NULL, NULL) == noErr)
			printf("Profile: %s\n", aName);

	} else {
		CMCloseProfile(rgbRef);
		if(aName) free(aName);
		return 1;
	}
	if(aName) free(aName);
	
	// get gammas
	
	// get rTRC
	if (CMGetProfileElement(rgbRef, cmRedTRCTag, &size, NULL) == noErr) {
		dataP = malloc(size);
		if (!dataP) {
			CMCloseProfile(rgbRef);
			return 1;
		}
		if (CMGetProfileElement(rgbRef, cmRedTRCTag, &size, dataP) != noErr) {
			free(dataP);
			CMCloseProfile(rgbRef);
			return 1;
		}
		count = CFSwapInt32BigToHost(((CMCurveType *)dataP)->countValue);
		if(count == 0) {
			// gamma 1
			Rgamma = 1.0;
			printf("Rgamma:  %f\n", Rgamma);
		} else if(count == 1) {
			// single u8Fixed8Number gamma
			dataPP = (unsigned char *)((CMCurveType *)dataP)->data;
			Rgamma = (float)(((int)dataPP[0]*256) + dataPP[1])/256.0;
			printf("Rgamma:  %f\n", Rgamma);
		} else {
			// ingore TRC profiles
			free(dataP);
			CMCloseProfile(rgbRef);
			printf("Can't handle TRC profiles\n");
			return 1;
		}
	}

	free(dataP);
	
	// get gTRC
	if (CMGetProfileElement(rgbRef, cmGreenTRCTag, &size, NULL) == noErr) {
		dataP = malloc(size);
		if (!dataP) {
			CMCloseProfile(rgbRef);
			return 1;
		}
		if (CMGetProfileElement(rgbRef, cmGreenTRCTag, &size, dataP) != noErr) {
			free(dataP);
			CMCloseProfile(rgbRef);
			return 1;
		}
		count = CFSwapInt32BigToHost(((CMCurveType *)dataP)->countValue);
		if(count == 0) {
			// gamma 1
			Ggamma = 1.0;
			printf("Ggamma:  %f\n", Ggamma);
		} else if(count == 1) {
			// single u8Fixed8Number gamma
			dataPP = (unsigned char *)((CMCurveType *)dataP)->data;
			Ggamma = (float)(((int)dataPP[0]*256) + dataPP[1])/256.0;
			printf("Ggamma:  %f\n", Ggamma);
		} else {
			// ingore TRC profiles
			free(dataP);
			CMCloseProfile(rgbRef);
			printf("Can't handle TRC profiles\n");
			return 1;
		}
	}
	
	free(dataP);
	
	// get bTRC
	if (CMGetProfileElement(rgbRef, cmBlueTRCTag, &size, NULL) == noErr) {
		dataP = malloc(size);
		if (!dataP) {
			CMCloseProfile(rgbRef);
			return 1;
		}
		if (CMGetProfileElement(rgbRef, cmBlueTRCTag, &size, dataP) != noErr) {
			CMCloseProfile(rgbRef);
			return 1;
		}
		count = CFSwapInt32BigToHost(((CMCurveType *)dataP)->countValue);
		if(count == 0) {
			// gamma 1
			Bgamma = 1.0;
			printf("Bgamma:  %f\n", Bgamma);
		} else if(count == 1) {
			// single u8Fixed8Number gamma
			dataPP = (unsigned char *)((CMCurveType *)dataP)->data;
			Bgamma = (float)(((int)dataPP[0]*256) + dataPP[1])/256.0;
			printf("Bgamma:  %f\n", Bgamma);
		} else {
			// ingore TRC profiles
			CMCloseProfile(rgbRef);
			printf("Can't handle TRC profiles\n");
			return 1;
		}
	}
	
	free(dataP);
	
	// get colorants as matrix A
	
	if (CMGetProfileElement(rgbRef, cmRedColorantTag, NULL, &fixedxyz) == noErr) {
		a[0]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].X));
		a[1]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Y));
		a[2]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Z));
		printf("Red:     X: %f, Y: %f, Z: %f\n", a[0], a[1], a[2]);
	} else {
		CMCloseProfile(rgbRef);
		return 1;
	}

	if (CMGetProfileElement(rgbRef, cmGreenColorantTag, NULL, &fixedxyz) == noErr) {
		a[3]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].X));
		a[4]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Y));
		a[5]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Z));
		printf("Green:   X: %f, Y: %f, Z: %f\n", a[3], a[4], a[5]);
	} else {
		CMCloseProfile(rgbRef);
		return 1;
	}

	if (CMGetProfileElement(rgbRef, cmBlueColorantTag, NULL, &fixedxyz) == noErr) {
		a[6]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].X));
		a[7]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Y));
		a[8]=FixedToFloat(CFSwapInt32BigToHost(fixedxyz.XYZ[0].Z));
		printf("Blue:    X: %f, Y: %f, Z: %f\n", a[6], a[7], a[7]);
	} else {
		CMCloseProfile(rgbRef);
		return 1;
	}

	// copy matrix A
	memcpy(a2, a, m*n*sizeof(float));
	
	//  compute LU factorization with clapack
	sgetrf_(&m, &n, a, &lda, ipiv, &info);
	if (info !=0) {
		CMCloseProfile(rgbRef);
		return info;
	}

	// setup colorWorlds
	
	// Lab profile ref
	if (CMGetDefaultProfileBySpace (cmLabData, &labRef) != noErr) {
		CMCloseProfile(rgbRef);
		CMCloseProfile(labRef);
		return 1;
	}
	
	// Lab -> RGB colorWorld
	if (NCWNewColorWorld (&cwRGBref, labRef, rgbRef) != noErr) {
		CMCloseProfile(rgbRef);
		CMCloseProfile(labRef);
		CWDisposeColorWorld (cwRGBref);
		return 1;
	};
	
	// XYZ profile ref
	if (CMGetDefaultProfileBySpace (cmXYZData, &xyzRef) != noErr) {
		CMCloseProfile(rgbRef);
		CMCloseProfile(labRef);
		CMCloseProfile(xyzRef);
		return 1;
	}

	// Lab -> XYZ colorWorld
	if (NCWNewColorWorld (&cwXYZref, xyzRef, rgbRef) != noErr) {
		CMCloseProfile(rgbRef);
		CMCloseProfile(labRef);
		CMCloseProfile(xyzRef);
		CWDisposeColorWorld (cwXYZref);
		return 1;
	};
	
	// close profiles
	CMCloseProfile(rgbRef);
	CMCloseProfile(labRef);
	CMCloseProfile(xyzRef);
	
	// input rgb list
	float		stdrgb[22*3] = {
		1.0, 0.0, 0.0,
		0.0, 1.0, 0.0,
		0.0, 0.0, 1.0,
		0.0, 1.0, 1.0,
		1.0, 0.0, 1.0,
		1.0, 1.0, 0.0,
		1.0, 1.0, 1.0,
		0.0, 0.0, 0.0,
		0.5, 0.0, 0.0,
		0.0, 0.5, 0.0,
		0.0, 0.0, 0.5,
		0.0, 0.5, 0.5,
		0.5, 0.0, 0.5,
		0.5, 0.5, 0.0,
		0.5, 0.5, 0.5,
		1.0, 0.5, 0.5,
		0.5, 1.0, 0.5,
		0.5, 0.5, 1.0,
		0.5, 1.0, 1.0,
		1.0, 0.5, 1.0,
		1.0, 1.0, 0.5,
		0.5, 0.5, 0.5
	};
	
	// copy rgb list
	float stdrgb2[22*3];
	memcpy(stdrgb2, stdrgb, 22*3*sizeof(float));
		

	// xyz/lab FloatBitmap for input list
	float	stdlab[22*3];
	CMFloatBitmap map_xyz_std = { 0,	// version
		{stdlab, stdlab+1, stdlab+2},	// base addrs
		1, 22,							// height, width
		22*3,							// rowStride
		3,								// colStride
		cmXYZData,
		kCMFloatBitmapFlagsNone};

	CMFloatBitmap map_lab_std = { 0,	// version
		{stdlab, stdlab+1, stdlab+2},	// base addrs
		1, 22,							// height, width
		22*3,							// rowStride
		3,								// colStride
		cmLabData,
		kCMFloatBitmapFlagsNone};

	// convert input list rgb->xyz
	for(i=0; i < 22*3; i += 3) {
		stdrgb[i+0] = pow(stdrgb[i+0],Rgamma);
		stdrgb[i+1] = pow(stdrgb[i+1],Ggamma);
		stdrgb[i+2] = pow(stdrgb[i+2],Bgamma);
		stdlab[i+0] = stdrgb[i+0]*a2[0]+stdrgb[i+1]*a2[3]+stdrgb[i+2]*a2[6];
		stdlab[i+1] = stdrgb[i+0]*a2[1]+stdrgb[i+1]*a2[4]+stdrgb[i+2]*a2[7];
		stdlab[i+2] = stdrgb[i+0]*a2[2]+stdrgb[i+1]*a2[5]+stdrgb[i+2]*a2[8];
	};
	
	// convert input list xyz->Lab
	if (CMConvertXYZFloatBitmap( &map_xyz_std, kCMIlluminantD50, 
			&map_lab_std, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
		if (cwRGBref) CWDisposeColorWorld (cwRGBref);
		if (cwXYZref) CWDisposeColorWorld (cwXYZref);
		return 1;
	}
	
	// test loop
	j=0;
	while(1){
		if(j < 22*3) {
			// process input list
			printf("\n---------------------------------------------------------------------------\n\n");
			printf("RGB input:                            ");
			printf("R: % 8.5f, G: % 8.5f, B: % 8.5f\n", stdrgb2[0+j], stdrgb2[1+j], stdrgb2[2+j]);
			lab[0]=stdlab[0+j];lab[1]=stdlab[1+j];lab[2]=stdlab[2+j];
			j+=3;
		} else {
			break;
			/*
			// get new Lab from user when done with input list
			printf("\ninput L*, a* and b* (\"-1 0 0\" to exit):\n");
			scanf("%f %f %f", &lab[0], &lab[1], &lab[2]);
			if (lab[0] < 0)  {
				if (cwRGBref) CWDisposeColorWorld (cwRGBref);
				if (cwXYZref) CWDisposeColorWorld (cwXYZref);
				return 0;
			}
			*/
		}
	
		printf("Lab input:                            ");
		printf("L: % 8.2f, a: % 8.2f, b: % 8.2f\n\n", lab[0], lab[1], lab[2]);

		// convert Bitmap Lab -> XYZ
		if (CMConvertXYZFloatBitmap( &map_lab, kCMIlluminantD50, 
				&map_xyz, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}
		
		// -----------   1.  ------------
		// solve system using clapack
		memcpy(b,xyz,3*sizeof(__CLPK_real));
		trans = 'N';
		sgetrs_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
		if (info !=0) return info;

		clipped = 0;
		for(i = 0; i<3; i++) {
			if(b[i]<0) {
				clipped |= 1;
				b[i] = 0;
			} else if(b[i]>1) {
				clipped |= 1;
				b[i] = 1;
			};
		}
				
		printf("LAPACK result");
		if(clipped) 
			printf(" (clipped):              ");
		else 
			printf(":                        ");
		
		printf("R: % 8.5f, G: % 8.5f, B: % 8.5f\n", 
			rgb[0] = pow(b[0],(1/Rgamma)), 
			rgb[1] = pow(b[1],(1/Ggamma)), 
			rgb[2] = pow(b[2],(1/Bgamma)));
			
		// roundtrip
		rgb[0]=pow(rgb[0], Rgamma); rgb[1]=pow(rgb[1], Ggamma); rgb[2]=pow(rgb[2], Bgamma); 
		
		xyz2[0] = rgb[0]*a2[0]+rgb[1]*a2[3]+rgb[2]*a2[6];
		xyz2[1] = rgb[0]*a2[1]+rgb[1]*a2[4]+rgb[2]*a2[7];
		xyz2[2] = rgb[0]*a2[2]+rgb[1]*a2[5]+rgb[2]*a2[8];
					
		// convert Bitmap XYZ -> Lab
		if (CMConvertXYZFloatBitmap( &map_xyz2, kCMIlluminantD50, 
				&map_lab2, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("roundtrip:                            L: % 8.2f, a: % 8.2f, b: % 8.2f\n", 
			lab2[0], lab2[1], lab2[2]);
			
		printf("                                Delta E: % 8.2f\n\n",
			sqrt(pow(lab2[0]-lab[0],2)+pow(lab2[1]-lab[1],2)+pow(lab2[2]-lab[2],2)));

		// -----------   2.  ------------
		// match Bitmap Lab -> RGB
		if (CMMatchFloatBitmap(cwRGBref,&map_lab,&map_rgb) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("CMMatchFloatBitmap Lab -> RGB result: ");
		printf("R: % 8.5f, G: % 8.5f, B: % 8.5f\n", rgb[0], rgb[1], rgb[2]);

		// roundtrip
		rgb[0]=pow(rgb[0], Rgamma); rgb[1]=pow(rgb[1], Ggamma); rgb[2]=pow(rgb[2], Bgamma); 
		
		xyz2[0] = rgb[0]*a2[0]+rgb[1]*a2[3]+rgb[2]*a2[6];
		xyz2[1] = rgb[0]*a2[1]+rgb[1]*a2[4]+rgb[2]*a2[7];
		xyz2[2] = rgb[0]*a2[2]+rgb[1]*a2[5]+rgb[2]*a2[8];
					
		// convert Bitmap XYZ -> Lab
		if (CMConvertXYZFloatBitmap( &map_xyz2, kCMIlluminantD50, 
				&map_lab2, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("roundtrip:                            L: % 8.2f, a: % 8.2f, b: % 8.2f\n", 
			lab2[0], lab2[1], lab2[2]);
			
		printf("                                Delta E: % 8.2f\n\n",
			sqrt(pow(lab2[0]-lab[0],2)+pow(lab2[1]-lab[1],2)+pow(lab2[2]-lab[2],2)));
		
		// -----------   3.  ------------
		// match fixed colors Lab -> RGB
		
		memcpy(labc, lab, 3*sizeof(float));
		clippedLab = 0;
	
		if(labc[0] < 0.0) {
			clippedLab |=1;
			labc[0] = 0.0;
		} else if (labc[0] > 100.0) {
			clippedLab |=1;
			labc[0] = 100.0;
		} 
		if(labc[1] < -128.0) {
			clippedLab |=1;
			labc[1] = -128.0;
		} else if (labc[1] > 127.996) {
			clippedLab |=1;
			labc[1] = 127.996;
		} 
		if(labc[2] < -128.0) {
			clippedLab |=1;
			labc[2] = -128.0;
		} else if (labc[2] > 127.996) {
			clippedLab |=1;
			labc[2] = 127.996;
		} 

		if (clippedLab) {
			printf("Lab clipped to UInt16:                ");
			printf("L: % 8.2f, a: % 8.2f, b: % 8.2f\n", 
				labc[0], labc[1], labc[2]
			);
		}
		
		fixedColor[0].Lab.L = (UInt16) (labc[0] * 655.35);
		fixedColor[0].Lab.a = (UInt16)((labc[1] + 128.0) * 256);
		fixedColor[0].Lab.b = (UInt16)((labc[2] + 128.0) * 256);

		if(CWMatchColors(cwRGBref, fixedColor, 1) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("CWMatchColors Lab -> RGB result:      ");
		printf("R: % 8.5f, G: % 8.5f, B: % 8.5f\n", 
			rgb[0] = (float)(fixedColor[0].rgb.red)/65535.0, 
			rgb[1] = (float)(fixedColor[0].rgb.green)/65535.0, 
			rgb[2] = (float)(fixedColor[0].rgb.blue)/65535.0
		);

		// roundtrip
		rgb[0]=pow(rgb[0], Rgamma); rgb[1]=pow(rgb[1], Ggamma); rgb[2]=pow(rgb[2], Bgamma); 

		xyz2[0] = rgb[0]*a2[0]+rgb[1]*a2[3]+rgb[2]*a2[6];
		xyz2[1] = rgb[0]*a2[1]+rgb[1]*a2[4]+rgb[2]*a2[7];
		xyz2[2] = rgb[0]*a2[2]+rgb[1]*a2[5]+rgb[2]*a2[8];
					
		// convert Bitmap XYZ -> Lab
		if (CMConvertXYZFloatBitmap( &map_xyz2, kCMIlluminantD50, 
				&map_lab2, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("roundtrip:                            L: % 8.2f, a: % 8.2f, b: % 8.2f\n", 
			lab2[0], lab2[1], lab2[2]);
			
		printf("                                Delta E: % 8.2f\n",
			sqrt(pow(lab2[0]-lab[0],2)+pow(lab2[1]-lab[1],2)+pow(lab2[2]-lab[2],2)));
		if (clippedLab) {
			printf("                 Delta E to clipped Lab: % 8.2f\n\n",
				sqrt(pow(lab2[0]-labc[0],2)+pow(lab2[1]-labc[1],2)+pow(lab2[2]-labc[2],2)));
		} else {
			printf("\n");
		}
			
		// -----------   4.  ------------
		// match Bitmap XYZ -> RGB
		if (CMMatchFloatBitmap(cwXYZref,&map_xyz,&map_rgb)  != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("CMMatchFloatBitmap XYZ -> RGB result: ");
		printf("R: % 8.5f, G: % 8.5f, B: % 8.5f\n", rgb[0], rgb[1], rgb[2]);
		
		// roundtrip
		rgb[0]=pow(rgb[0], Rgamma); rgb[1]=pow(rgb[1], Ggamma); rgb[2]=pow(rgb[2], Bgamma); 

		xyz2[0] = rgb[0]*a2[0]+rgb[1]*a2[3]+rgb[2]*a2[6];
		xyz2[1] = rgb[0]*a2[1]+rgb[1]*a2[4]+rgb[2]*a2[7];
		xyz2[2] = rgb[0]*a2[2]+rgb[1]*a2[5]+rgb[2]*a2[8];
					
		// convert Bitmap XYZ -> Lab
		if (CMConvertXYZFloatBitmap( &map_xyz2, kCMIlluminantD50, 
				&map_lab2, kCMIlluminantD50, cmUseDefaultChromaticAdaptation) != noErr) {
			if (cwRGBref) CWDisposeColorWorld (cwRGBref);
			if (cwXYZref) CWDisposeColorWorld (cwXYZref);
			return 1;
		}

		printf("roundtrip:                            L: % 8.2f, a: % 8.2f, b: % 8.2f\n", 
			lab2[0], lab2[1], lab2[2]);
			
		printf("                                Delta E: % 8.2f\n",
			sqrt(pow(lab2[0]-lab[0],2)+pow(lab2[1]-lab[1],2)+pow(lab2[2]-lab[2],2)));
			
		lab[0] = -1.0;
	}
	
	if (cwRGBref) CWDisposeColorWorld (cwRGBref);
	if (cwXYZref) CWDisposeColorWorld (cwXYZref);
	return 0;
}
